//-----------------------------------------------------------------------------
// gestAnello.java
// package: FileServer
// 
// gestore di una rete con una topologia ad anello
//
// data: 16-01-2000
//-----------------------------------------------------------------------------

package FileServer;

import java.net.*;
import java.io.*;
import java.util.*;

public class gestAnello extends gestore {

   Vector coda;    //coda dei messaggi (in attesa del ripristino della connessione)
   InetAddress succAdd;  //IP del successore
   int         succPort; //porta di ingresso del successore
   boolean disp;      //true se la connessione con il successore  valida
   InetAddress succ2Add;  //IP del successore del successore
   int         succ2Port; //porta di ingresso del successore del successore
   InetAddress pred;   //IP del predecessore
   public static long delay=4000;//ritardo tra 2 tentativi di connessione( in msec )
   boolean     first;   //true se  la prima segnalazione di un crash out
   public static int  inPortDefault = 2000;
   int         inPort; //porta cui collegare le socket di ingresso (TCP)
   public static int  outPortDefault = 2001;
   int         outPort; //porta cui collegare le socket di uscita (TCP)
   public static int  portClientDefault = 2000; 
   int         portClient; //porta cui  collegato l'interprete cliente (UDP)
   String      fileConfig = "cfg.gs";
   
   public gestAnello() throws Exception {
      //creazione di un nameServer vuoto
      ns = new nameServer(this);
      init();
      coda = new Vector(0,5);   //crea coda messaggi (vuota)
      inPort = gestAnello.inPortDefault;
      outPort = gestAnello.outPortDefault;
      portClient = portClientDefault;
      iC = new intClient( this );
      //creazione del thread di ascolto per connessioni server della rete
      new initInSock( this );
   }//gestAnello
   
   public gestAnello( String fileNameServer ) throws Exception {
      ns = new nameServer( fileNameServer, this );
      init();
      coda = new Vector(0,5);
      inPort = inPortDefault;
      outPort = outPortDefault;
      portClient = portClientDefault;
      iC = new intClient( this );
      //creazione del thread di ascolto per connessioni server della rete
      new initInSock( this );
   }//gestAnello
   
   public gestAnello( String fileNameServer, String fileConfig ) throws Exception {
      ns = new nameServer( fileNameServer, this );
      this.fileConfig = fileConfig;
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello): creato thread ascolto connessioni ingresso");
      init();
      coda = new Vector(0,5);
      //esecuzione della procedura di ripristino delle socket di uscita, delle 
      //porte di ingresso ed uscita
      loadConnOut( fileConfig );
      iC = new intClient( this );
   }//gestAnello
   
   public gestAnello( int inPort, int outPort, int portClient ) throws Exception {
      //creazione di un nameServer vuoto
      ns = new nameServer(this);
      init();
      coda = new Vector(0,5);   //crea coda messaggi (vuota)
      this.inPort = inPort;
      this.outPort = outPort;
      this.portClient = portClient;
      iC = new intClient( this );
      //creazione del thread di ascolto per connessioni server della rete
      new initInSock( this );
   }//gestAnello

   public gestAnello( int inPort, int outPort ) throws Exception {
      this( inPort, outPort, portClientDefault );
   }//gestAnello
   
   public gestAnello( String fileNameServer, int inPort, int outPort, int portClient ) 
            throws Exception {
      //creazione di un nameServer vuoto
      ns = new nameServer( fileNameServer, this );
      init();
      coda = new Vector(0,5);
      this.inPort = inPort;
      this.outPort = outPort;
      this.portClient = portClient;
      iC = new intClient( this );
      //creazione del thread di ascolto per connessioni server della rete
      new initInSock( this );
   }//gestAnello
 
   public gestAnello( String fileName, int inPort, int outPort ) throws Exception {
      this( fileName, inPort, outPort, portClientDefault );
   }//gestAnello
   
   protected void init() {
      //impostazione della tabella dei buffer di uscita (crea nuova istanza di trasm)
      tr = new trasm( this, 1, 0 );
      //impostazione della tabella dei buffer di ingresso ( "        "   di intServer)
      is = new intServer( this, ns, tr, 1, 0 );
      //creazione delle 2 tabelle per le socket di ingresso ed uscita
      inSock = new Vector( 1, 1 );
      outSock = new Vector( 1, 0 );
   }//init

   
   protected void loadConnOut( String fileConfig ) throws Exception {
   /* metodo per il ripristino delle tabelle del nodo
      Struttura del file di configurazione:
      ( inPort, outPort )
      ( InetAddress, porta ) successore
      ( InetAddress, porta ) successore del successore
      ( InetAddress )        predecessore
      ( portClient )
   */
      if (DbFS.gestAnello)
         DbFS.outMexLn("(gestAnello) (loadConnOut): inizio deserializzazione");
      ObjectInputStream obj = new ObjectInputStream ( new FileInputStream( fileConfig ));
      inPort = ( (Integer) obj.readObject() ).intValue();
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (loadConnOut): letto inPort "
         + String.valueOf( inPort ) );
      //creazione del thread di ascolto per connessioni server della rete
      new initInSock( this );
      outPort = ( (Integer) obj.readObject() ).intValue();
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (loadConnOut): letto outPort"
         + String.valueOf( outPort ) );
      succAdd = (InetAddress) obj.readObject();
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (loadConnOut): letto InetAddress successore "
         + succAdd.toString() );
      succPort = ( (Integer) obj.readObject() ).intValue();
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (loadConnOut): letta porta successore "
         + String.valueOf( succPort ) );
      succ2Add = (InetAddress) obj.readObject();
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (loadConnOut): letto InetAddress successore2 "
         + succ2Add.toString() );
      succ2Port = ( (Integer) obj.readObject() ).intValue();
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (loadConnOut): letta porta successore2 "
         + String.valueOf( succ2Port ) );
      pred = (InetAddress) obj.readObject();
      if (DbFS.gestAnello)
         DbFS.outMexLn("(gestAnello) (loadConnOut): letto InetAddress predecessore "
         + pred.toString() );
      portClient = ( (Integer) obj.readObject() ).intValue();
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (loadConnOut): letta porta intClient "
         + String.valueOf( portClient ) );
      obj.close();
      //ripristino del successore
      ripristino();
   }//loadConnOut
   
   void ripristino( ) {
      //ripristino successore( nell'hp che sia presente )
      Socket succ = ripSucc( succAdd, succPort );
      String sinc = new String(); //Oggetto che uso per la wait
      while( succ == null ) {
         //inserisco un ritardo di gestAnello.delay msec
         succ = ripSucc( succAdd, succPort );
         try {
            if (DbFS.gestAnello) 
               DbFS.outMexLn("(gestAnello)(ripristino): attesa di " + 
                  String.valueOf( gestAnello.delay/1000 ) + " secondi" );
            synchronized( sinc ) {
               sinc.wait(gestAnello.delay);
            }
         } catch ( Exception e ) { }
      }
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello)(ripristino): ripristinato successore" +
         " (succ = " + succ.toString() + ")" );
      try {
      succ.setTcpNoDelay( true );
      if (DbFS.gestAnello)
         DbFS.outMexLn("(gestAnello)(ripristino): succ TcpNoDelay="+
         String.valueOf( succ.getTcpNoDelay() ) );
      } catch (Exception e) {}
      //inserimento nella tabella
      newOutSocket( succ );
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello)(ripristino): eseguita newOutSock");
      //aggiornamento successore
      disp = true;
      first = true;   
   }//ripristino
   
   Socket ripSucc( InetAddress addr, int port ) {
      return ripSucc( addr, port, outPort );
   }//ripSucc
   
   Socket ripSucc( InetAddress addr, int port, int localPort ) {
      try {
         Socket succ=null;
         if (localPort>0) {
            if (DbFS.gestAnello) 
               DbFS.outMexLn("(gestAnello)(ripSucc): tento di collegarmi con "+addr+
               " porta="+String.valueOf( port )+" dalla porta "+String.valueOf( localPort ) );
            succ = new Socket( addr, port, InetAddress.getLocalHost(), localPort );
         } else {
            succ = new Socket( addr, port );
         }
         return succ;
      } catch (Exception e) {
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello)(ripSucc): Eccezione "+e.toString() );
         return null;
      }
   }//ripSucc
   
   protected void saveConnOut( String fileConfig ) throws Exception {
      //metodo per il salvataggio della tabella delle socket in uscita
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (saveConnOut): salvataggio configurazione socket" );
      ObjectOutputStream obj = new ObjectOutputStream( new FileOutputStream( fileConfig ) );
      obj.writeObject( new Integer( inPort ) );
      obj.writeObject( new Integer( outPort ) );
      obj.writeObject( succAdd );
      obj.writeObject( new Integer( succPort ) );
      obj.writeObject( succ2Add );
      obj.writeObject( new Integer( succ2Port) );
      obj.writeObject( pred );
      obj.writeObject( new Integer( portClient ) );
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (saveConnOut): salvataggio OK" );
   }//saveConnOut

   public synchronized void newInSocket( Socket sIn ) {
      //inserimento della socket nella tabella delle socket di ingresso
      inSock.addElement( sIn );
      try {
         InputStream inStr = sIn.getInputStream();
         ObjectInputStream inObj = new ObjectInputStream( inStr );
         //inserimento nella tabella degli ObjectInputStream
         is.ins( inObj );
      } catch ( IOException e ) {
         //rimozione dell'entry relativa alla socket
         inSock.removeElement( sIn );
         if (DbFS.gestAnello) 
            DbFS.outMexLn("(gestAnello): (newInSocket) Eccezione " + e.toString() );
      }
   }//newInSocket

   public synchronized void newOutSocket( Socket sOut ) {
      //richiesta di inserimento di una nuova socket di uscita
      //inserimento della socket nella tabella delle socket di uscita
      if (DbFS.gestAnello) DbFS.outMexLn("(gestAnello) (newOutSocket): ingresso" );
      //inserimento nuova entry nella tabella delle socket di uscita
      outSock.addElement( sOut );
      try {
         //inserimento nella tabella di trasmissione
         OutputStream oStr = sOut.getOutputStream();
         ObjectOutputStream oStrObj = new ObjectOutputStream( oStr );
         tr.ins( oStrObj);
      } catch ( IOException e ) {
         //operazione fallita, cancello quanto gi fatto
         outSock.removeElement( sOut );
         if (DbFS.gestAnello) 
            DbFS.outMexLn("(gestAnello): (newOutSocket) Eccezione "+ e.toString() );
      }
   }//newOutSocket

   public synchronized void notifyCrashOut( int ind, Comando cmd ) {
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (notCrashOut): caduto ind " + String.valueOf( ind ));
      if (first) {
         first = false;
         if ( disp ) {
            //caduta del successore
            disp = false;
            //avvia thread per la gestione della caduta del successore
            if (DbFS.gestAnello) 
               DbFS.outMexLn("(gestAnello) (notCrashOut): caduto successore");
            new crashSucc( this );
         }
         else {
            if (DbFS.gestAnello) 
               DbFS.outMexLn("(gestAnello) (notCrashOut): successore tornato");
            //il successore  tornato
            //avvia thread per la gestione del ritorno
            new ritSucc( this );
         }
      }      
      coda.addElement( cmd ); //inserisci messaggio in coda
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (notifyCrashOut): inserito in coda " 
         + cmd.toString() );
   }//notifyCrashOut
   
   public synchronized void notifyCrashIn( int ind ) {
      if (DbFS.gestAnello) 
         DbFS.outMexLn("(gestAnello) (notCrashIn): caduto ind "+String.valueOf(ind) );
      //chiusura della socket
      try {
         ( (Socket) inSock.elementAt( ind ) ).close();
         if (DbFS.gestAnello)
           DbFS.outMexLn("(gestAnello) (notCrashIn): chiusa socket " +
           String.valueOf( ind ) );
      } catch ( Exception e ) { }
      //cancellazione dell'entry relativa nella tabella delle socket di in
      inSock.removeElementAt( ind );
      if (DbFS.gestAnello)
        DbFS.outMexLn("(gestAnello) (notCrashIn): cancellata entry socket " +
           String.valueOf( ind ) + "\ninSock: "+String.valueOf(inSock.size())+
           " elementi" );
      is.canc( ind );
      if (DbFS.gestAnello)
        DbFS.outMexLn("(gestAnello) (notCrashIn): cancellata entry ricezione  " +
        String.valueOf( ind ) + "\nNuova tabella ricezione: " + is.toString() );
   }//notifyCrashIn
   
   public synchronized void aggiornaOut() {
   //metodo per l'aggiornamento del successore
      if (DbFS.gestAnello)
        DbFS.outMexLn("(gestAnello)(aggiornaOut): aggiornamento successore");
   }//aggiornaOut
   
   public synchronized void aggiornaIn() {
   //metodo per l'aggiornamento dal predecessore
      if (DbFS.gestAnello)
        DbFS.outMexLn("(gestAnello)(aggiornaIn): aggiornamento da predecessore");
   }//aggiornaIn
     
   public static void main( String args[] ) {
      DbFS.OpenDbgWin("gestAnello");
      DbFS.CloseDbgWin();
      try {
         new gestAnello();
      } catch ( Exception e ) {}
   }//main 
   
}//gestAnello

class crashSucc extends Thread {
   //thread di gestione della caduta del successore
   
   gestAnello gs;
   
   public crashSucc( gestAnello gs ) {
      super();
      this.gs = gs;
      start();
   }//crashSucc

   public void run() {
      String sinc=new String(); //oggetto che uso per wait
      //ritenta il collegamento nel caso di caduta della linea e non del nodo
      if ( gs.outSock.size() > 0 ) {
         Socket s = (Socket) gs.outSock.elementAt( 0 );
         try { s.close(); } catch( Exception e ) { }
      }
      Socket succ = gs.ripSucc( gs.succAdd, gs.succPort );
      if ( succ == null ) {
         //non  possibile ripristinare il collegamento
         //per hp. il succ2 deve essere presente
         succ = gs.ripSucc( gs.succ2Add, gs.succ2Port );
         if (DbFS.crashSucc) 
            DbFS.outMexLn( "(crashSucc): tento di collegarmi con successore 2" + 
            "ind:" + gs.succ2Add.toString()+" porta:"+String.valueOf(gs.succ2Port) );
         while( succ == null ) {
            try {
               synchronized( sinc ) {
                  sinc.wait(gestAnello.delay);
               }
            } catch( Exception e ) {
               if (DbFS.crashSucc) 
                  DbFS.outMexLn("(crashSucc): interrotto wait("+e.toString()+")"); 
            }
            succ = gs.ripSucc( gs.succ2Add, gs.succ2Port );
         }
         if (DbFS.crashSucc) 
            DbFS.outMexLn( "(crashSucc): collegamento con successore 2 riuscito" );
      }
      //cancello la vecchia entry nella tabella di trasmissione
      //il controllo  fatto per contemplare l'ipotesi che questo sia il primo
      //collegamento ed il successore non sia mai stato raggiungibile (ci si
      //collega subito a succ2 )
      if ( gs.outSock.size() > 0 ) {
         gs.tr.canc( 0 );
         gs.outSock.removeElementAt( 0 );
      }
      gs.newOutSocket( succ );
      gs.first = true;
      //svuota coda dei messaggi in attesa
      if (DbFS.crashSucc) 
         DbFS.outMexLn("(crashSucc): la coda dei messaggi ne contiene " + 
         String.valueOf( gs.coda.size() ) );
      while ( gs.coda.size() > 0 ) {
         gs.tr.invia( (Comando) gs.coda.elementAt( 0 ) );
         gs.coda.removeElementAt( 0 );
      }
      if (DbFS.crashSucc) 
         DbFS.outMexLn("(crashSucc): svuotata coda dei messaggi" );
   }//run

}//crashSucc

class ritSucc extends Thread {
   //thread per la gestione del ritorno del successore caduto
   
   gestAnello gs;
   
   public ritSucc( gestAnello gs ) {
      super();
      this.gs = gs;
      start();
   }//ritSucc
   
   public void run() {
      //cancellazione collegamento con successore del successore
      try {
         ( (Socket) gs.outSock.elementAt( 0 ) ).close();
      } catch( Exception e ) { 
         if (DbFS.ritSucc) 
            DbFS.outMexLn("(ritSucc): Eccezione (chiusura sock succ2) "+e.toString() );
      }
      gs.tr.canc( 0 );   
      gs.outSock.removeElementAt( 0 );
      gs.ripristino();   
      //aggiornamento tabella successore
      gs.aggiornaOut();
      //svuota coda dei messaggi in attesa
      if (DbFS.ritSucc) 
         DbFS.outMexLn("(ritSucc): la coda dei messaggi ne contiene " + 
         String.valueOf( gs.coda.size() ) );
      while ( gs.coda.size() > 0 ) {
         gs.tr.invia( (Comando) gs.coda.elementAt( 0 ) );
         gs.coda.removeElementAt( 0 );
      }
      if (DbFS.ritSucc) 
         DbFS.outMexLn("(ritSucc): svuotata coda dei messaggi" );
   }//run
   
}//ritSucc