//-----------------------------------------------------------------------------
// gestReteAnello.java
// package: FileServer
// 
// modulo che si occupa dell'inserimento di nuovi nodi nella rete (con topologia
// ad anello)
//
// data: 27-02-2000
//-----------------------------------------------------------------------------

package FileServer;

import java.net.*;
import java.io.*;
import java.util.*;

public class gestReteAnello extends gestAnelloNSPatch{

   protected static int maxElem = 20;
   protected Vector richieste; //colonna richieste (contiene gli IP dei server)
   protected Vector passi;     //colonna passi 
   protected intNewServer iNs;  //interprete comandi (su porta UDP )
   public static int  portServerDefault = 8001;
   int portServer;    //porta cui  agganciato l'interprete server
   int portServPred;  // "      "          "         "              nel predecessore
   int portServPred2; // "      "          "         "              nel pred2        
   int portServSucc;  // "      "          "         "              nel successore
   int inPortNS; //porta di ingresso rete server comunicata dal nuovo nodo
   Socket  sockCom;         //socket da cui si ricevono i comandi
   
   public gestReteAnello( String fileName, boolean nuovo ) throws Exception {
      //nuovo = true per un nuovo server da inserirsi
      super( fileName );
      richieste = new Vector( maxElem, 0 );
      passi = new Vector( maxElem, 0 );
      portServer = portServPred = portServPred2 = portServSucc = portServerDefault;
      iNs = new intNewServer( this );
      if (!nuovo) new ascoltaRic( this );
   }//gestReteAnello

   public gestReteAnello( String fileNameServer, int inPort, int outPort, int portClient,
         int portServer, boolean nuovo ) throws Exception {
      super( fileNameServer, inPort, outPort, portClient );
      richieste = new Vector( maxElem, 0 );
      passi = new Vector( maxElem, 0 );
      this.portServer = portServer;
      iNs = new intNewServer( this );
      if (!nuovo) new ascoltaRic( this );
   }//gestReteAnello

   public gestReteAnello( String fileNameServer, int inPort, int outPort,
         int portServer, boolean nuovo ) throws Exception {
      this( fileNameServer, inPort, outPort, portClientDefault, portServer, nuovo );
   }//gestReteAnello
   
   public gestReteAnello( String fileNameServer, int inPort, int outPort, int portServer, 
         int portServPred, int portServPred2, int portServSucc ) 
         throws Exception {
      this( fileNameServer, inPort, outPort, portServer, false );
      this.portServPred = portServPred;
      this.portServPred2 = portServPred2;
      this.portServSucc = portServSucc;         
   }//gestReteAnello
   
   public gestReteAnello( String fileNameServer, String fileConfig, int portServer,
      int portServPred, int portServPred2, int portServSucc ) 
         throws Exception {
      super( fileNameServer, fileConfig );  
      this.portServer = portServer;
      this.portServPred = portServPred;
      this.portServPred2 = portServPred2;
      this.portServSucc = portServSucc;         
      richieste = new Vector( maxElem, 0 );
      passi = new Vector( maxElem, 0 );
      iNs = new intNewServer( this );
      new ascoltaRic( this );
   }//gestReteAnello
   
   public gestReteAnello( String fileNameServer, String fileConfig ) 
         throws Exception {
      this( fileNameServer, fileConfig, portServerDefault, portServerDefault,portServerDefault,
      portServerDefault );
   }//gestAnelloRete
   
   public void operServ( Socket sAgg, InetAddress newServer, int portNS ) 
         throws Exception {
      //metodo di aggiornamento (lato server gi in anello)
      ObjectOutputStream outObj = new ObjectOutputStream( sAgg.getOutputStream() );
      ObjectInputStream inObj = new ObjectInputStream( sAgg.getInputStream() );
      nameServerRid nsRid;
      Vector newRicTab;
      //invio disp
      outObj.writeBoolean( disp );
      outObj.flush();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito disp");
      boolean cont = inObj.readBoolean();
      if ( cont ) {
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: ricevuto assenso");
         boolean oldDisp;
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: disp="+String.valueOf(disp));
         synchronized( ns ) {
            //creazione copia fisica del name server
            nsRid = new nameServerRid( ns );
            oldDisp = disp;
            disp = false;
            //creazione copia fisica della tabella ricTab
            newRicTab = (Vector) ricTab.clone();
            if ( ricTab.size() > 0 ) ricTab.removeAllElements();
         }
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: disp="+String.valueOf(disp)+
            "; oldDisp="+String.valueOf(oldDisp));
         //invio name server ridotto
         outObj.writeObject( nsRid );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito nsRid");
         //invio newRicTab ( la tabella che il nuovo nodo deve inviare al suo successore
         outObj.writeObject( newRicTab );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito ricTab");
         //gR.coda ora  vuota
         Vector newCoda = (Vector) coda.clone();
         if ( oldDisp ) {
            //invio tabella comandi persi per successore di nuovo nodo
            outObj.writeObject( newCoda );            
            if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito newCoda="+newCoda);
         }
         else {
            //successore caduto (all'avvio dovr aggiornare pred, portServerPred)
            newCoda.addElement( new aggSucc( newServer, portNS ) );
            outObj.writeObject( newCoda );
            if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito newCoda="+newCoda);
         }
         //invio succ, succ2, pred, portServer, portServPred, portServSucc
         outObj.writeObject( succAdd );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito succAdd");
         outObj.writeInt( succPort );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito succPort");
         outObj.writeObject( succ2Add );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito succ2Add");
         outObj.writeInt( succ2Port );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito succ2Port");
         outObj.writeObject( InetAddress.getLocalHost() );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito localHost");
         outObj.writeInt( portServer );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito portServer");
         outObj.writeInt( portServPred );
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito portServPred");
         outObj.writeInt( portServSucc );
         outObj.flush();
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: spedito portServSucc\n"+
            "operServ: in attesa di inPortNS ");
         //ricezione inPort (per creazione canale) succPort
         inPortNS = inObj.readInt();
         if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: ricevuto inPortNS");
      }
      else
         if (DbFS.gestReteAnello)
            DbFS.outMexLn("operServ: il nuovo server si  gi inserito nell'anello");
   }//operServ
   
   public void operNewServ( ServerSocket ser, Socket sAgg ) throws Exception {
      //metodo di aggiornamento (lato server che vuole entrare in anello )
      ObjectOutputStream outObj = new ObjectOutputStream( sAgg.getOutputStream() );
      ObjectInputStream inObj = new ObjectInputStream( sAgg.getInputStream() );
      //ricezione disp
      disp = inObj.readBoolean();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto disp");
      outObj.writeBoolean( true );
      outObj.flush();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: inviato assenso a continuare");
      //ricezione name Server ridotto
      nameServerRid nsRid = (nameServerRid) inObj.readObject();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto nsRid");
      //aggiornamento name server attuale
      ns.fisico = nsRid.fisico;
      ns.logico = nsRid.logico;
      ns.nelem = nsRid.nelem;
      ns.salva();
      //ricezione tabella richieste per successore
      ricTab = (Vector) inObj.readObject();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto ricTab="+ricTab);
      //ricezione comandi persi per successore
      coda = (Vector) inObj.readObject();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto coda="+coda);
      //ricezione name server
      //ricezione succ, succ2, pred, portServPred, portServPred2, portServSucc
      succAdd = (InetAddress) inObj.readObject();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto succAdd");
      succPort = inObj.readInt();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto succPort");
      succ2Add = (InetAddress) inObj.readObject();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto succ2Add");
      succ2Port = inObj.readInt();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto succ2Port");
      pred = (InetAddress) inObj.readObject();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto pred");
      portServPred = inObj.readInt();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto portServPred");
      portServPred2 = inObj.readInt();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto portServPred2");
      portServSucc = inObj.readInt();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: ricevuto portServSucc");
      //invio inPort
      outObj.writeInt( inPort );
      outObj.flush();
      if (DbFS.gestReteAnello) DbFS.outMexLn("operNewServ: spedito inPort\n" +
         "operNewServ: attesa ack per avvio ripristino" );
      //attesa ack
      try {
         int ack = inObj.readInt();
      } catch (Exception e) { }
      if (DbFS.gestReteAnello) 
         DbFS.outMexLn("operNewServ: ricevuto ack, avvio ripristino" );
      if (disp){
         ripristino();
         if (DbFS.gestReteAnello) 
            DbFS.outMexLn("operNewServ: ripristino successore" );
      }
      else {
         //successore non presente, prima di creare collegamento con succ2 devo togliere
         //il comando di aggiornamento dalla coda (o i comandi di aggiornamento)
         Vector codaTemp = new Vector( 1, 5 );
         boolean esci = false;
         while ( coda.size() > 0 && !esci ) {
            Comando cmd = (Comando) coda.elementAt( 0 );
            if ( ((cmd instanceof aggPred) || (cmd instanceof aggSucc)) ) {
               codaTemp.addElement( cmd );
               coda.removeElementAt( 0 );
            }
            else esci = true;
         }
         if (DbFS.gestReteAnello) 
            DbFS.outMexLn("operNewServ: tolti da coda i comandi di aggiornamento; coda="+
            coda );
         crashSucc cS = new crashSucc( this );
         if (DbFS.gestReteAnello) 
            DbFS.outMexLn("operNewServ: avviato ripristino successore2, attesa terminazione" );
         try {
            cS.join();
         } catch( Exception e ) {
            if (DbFS.gestReteAnello) 
               DbFS.outMexLn("operNewServ: Eccezione join "+e );
         }
         if (DbFS.gestReteAnello) 
            DbFS.outMexLn("operNewServ: ripristino successore2 terminato" );
         //ripristino tabella (ora  vuota)
         coda = codaTemp;
         if (DbFS.gestReteAnello) 
            DbFS.outMexLn("operNewServ: coda "+coda);
      }
      if (DbFS.gestReteAnello) DbFS.outMexLn("operServ: salvataggio configurazione ("+
         ns.name+".gs)" );
      saveConnOut( ns.name + ".gs" );
      //chiusura socket ricezione comando
      sAgg.close();
      //avvia thread ascolto comandi di modifica rete
      new ascoltaRic( this, ser );
    }//operNewServ
    
    public int contaPassi( InetAddress addNS ) {
      int passi = 999;
      try {
         String cmd[] = new String[3];
         cmd[0] = new String( "tracert" );
         cmd[1] = new String( "-d" );
         cmd[2] = new String( addNS.getHostName() );
         Runtime run = Runtime.getRuntime();
         Process dir = run.exec( cmd  );
         BufferedReader b = new BufferedReader( new InputStreamReader ( 
            dir.getInputStream() ) );
         String line = b.readLine();
         boolean esci = false;
         while ( !esci ) {
            if (line!=null) {
               StreamTokenizer st = new StreamTokenizer( new StringReader( line ) );
               st.nextToken();
               if ( st.ttype == StreamTokenizer.TT_NUMBER ) passi = (int) st.nval;
               line = b.readLine();
            }
            else esci = true;
         }
      } catch( Exception e) {
            System.out.println("Eccezione Thread: "+e.toString() );
      }
      finally {
         return passi;
      }
    }//contaPassi

}//gestReteAnello

class ascoltaRic extends Thread {
//thread per l'accettazione di nuove connessioni e la ricezione dei comandi
   gestReteAnello gR;
   boolean nuovo;
   ServerSocket ser;

   public ascoltaRic(gestReteAnello gR) {
      super();
      this.gR = gR;
      this.nuovo = false;
      start();
   }//ascoltaRic
   
   public ascoltaRic(gestReteAnello gR, ServerSocket ser ){
      super();
      this.gR = gR;
      this.nuovo = true;
      this.ser = ser;
      start();
   }//ascoltaRic
   
   ServerSocket newServ( int port, boolean nuovo ) {
      if ( !nuovo ) {
         String sinc = "";
         try {
            ServerSocket serN = new ServerSocket( port );
            return serN;
         } catch (Exception e) {
            try {
               synchronized( sinc ) {
               if (DbFS.ascoltaRic) DbFS.outMexLn("ascoltaRic: attesa 5 secondi" );
                  sinc.wait( 5000 );
               }
            }
            catch (Exception exc) {
            } 
            finally {
               return newServ( port, false );
            }
         }
      }
      else { //nuovo nodo
         return ser;
      }
   }//newServ
   
   public void run(){
      try {
         ServerSocket ser = newServ( gR.portServer, nuovo );
         while ( true ) {
            try {
               if (DbFS.ascoltaRic) DbFS.outMexLn("ascoltaRic: in attesa di connessione "+
                  "alla porta: " + String.valueOf( gR.portServer ) );
               gR.sockCom = ser.accept();
               if (DbFS.ascoltaRic) DbFS.outMexLn("ascoltaRic: accettata connessione");
               gR.sockCom.setSoLinger( true, 100 );
               gR.sockCom.setTcpNoDelay( true ); 
               ObjectInputStream socObj = new ObjectInputStream( gR.sockCom.getInputStream() );
               try {
                  while (true ) {
                     Comando cR = (Comando) socObj.readObject();
                     if (DbFS.ascoltaRic) 
                        DbFS.outMexLn("ascoltaRic: ricevuto comando cR="+cR);
                     cR.set( gR );
                     cR.start();
                     if (DbFS.ascoltaRic) 
                       DbFS.outMexLn("ascoltaRic: avviato cR="+cR);
                  }
               } catch ( Exception e ) {
                  if ( e instanceof ClassCastException ) {
                     //controllo di ricezione comando o qualcosa di diverso
                     if (DbFS.ascoltaRic) 
                        DbFS.outMexLn("ascoltaRic: si  ricevuto qualcosa di diverso"+
                        " da un comando" );
                     //creazione risposta
                     ObjectOutputStream outObj = new ObjectOutputStream(
                        gR.sockCom.getOutputStream() );
                     outObj.writeBoolean( false );
                     outObj.flush();
                     outObj.close();
                  }
                  else 
                     if (DbFS.ascoltaRic) 
                        DbFS.outMexLn("ascoltaRic: Eccezione "+e.toString() );
                  //chiuso estremo di  invio
               }
            }
            catch( Exception e ) {
              if (DbFS.ascoltaRic) DbFS.outMexLn("ascoltaRic: Eccezione "+e.toString() );
              //accept
            }
         }
      } catch( Exception e ) {
         if (DbFS.ascoltaRic) DbFS.outMexLn("ascoltaRic: Eccezione "+e.toString() );
         //new ServerSocket
      }
   }//run
   
}//ascoltaRic