//-----------------------------------------------------------------------------
// nameServer.java
// package: FileServer
// 
// gestore del servizio di nomi ( locale )
//
// data: 27-12-99
//       24-01-2000 inseriti metodi di salvataggio e ripristino
//       20-02-2000 inseriti log e prop
//
//-----------------------------------------------------------------------------

package FileServer;

import java.util.*;
import java.io.*;
import java.net.*;

public class nameServer implements Serializable {

   protected gestore gs;      //gestore cui  collegato questo name server
	protected Vector fisico;   //colonna nomi fisici (tabella corrispondenze)
	protected Vector logico;   //colonna nomi logici (tabella corrispondenze)
	protected Vector prop;     //tabella delle proposte del nodo
	protected static String StringNull = new String( "" );
	protected static String fisicoEsterno = new String( "**ext**" );
	protected int nelem = 0; 		//numero path validi nella tabella corrispondenze
	protected static int dimDefault = 50;  //dimensione di default tabelle
	protected static int incDefault = 5;   //incremento di default tabelle
	
	protected boolean risc = false; //true durante il "riscaldamento" della copia fredda
	//files
	protected static String nameDefault =  "cfg";
	protected static String extNs =  ".ns";
	protected static String extLog = ".log";
	protected static String extProp =".prp";
	protected RandomAccessFile fileLogRW;
	protected RandomAccessFile filePropRW;
	protected FileOutputStream fileLog;
	protected FileOutputStream fileProp;
	protected ObjectOutputStream logObj;
	protected ObjectOutputStream propObj;
	protected String name = nameDefault;

   public nameServer( ) {
      this( (gestore) null );
   }//nameServer

   public nameServer( gestore gs ) {
   //creazione di un nameServer vuoto
      this.gs = gs;
      fisico = new Vector( nameServer.dimDefault, nameServer.dimDefault );
      logico = new Vector( nameServer.dimDefault, nameServer.dimDefault );
      prop = new Vector( nameServer.dimDefault, nameServer.dimDefault );
      //riscaldo nuova copia name server
      riscalda();
      try {
         initFile();
      } catch( Exception e ) {
         if (DbFS.nameServer) 
            DbFS.outMexLn("(nameServer) Eccezione apertura file in scrittura "+e.toString());
      } 
  }//nameServer
   
	public nameServer( int dim, int inc, gestore gs ) {
	//creazione di un nameServer vuoto
	   this.gs = gs;
	   if ( dim > 0 && inc > 0 ) {
	      fisico = new Vector( dim, inc );
	 	   logico = new Vector( dim, inc );
	 	   prop = new Vector( dim, inc );
		}
		else {
         fisico = new Vector( nameServer.dimDefault, nameServer.dimDefault );
         logico = new Vector( nameServer.dimDefault, nameServer.dimDefault );
         prop = new Vector( nameServer.dimDefault, nameServer.dimDefault ); 
      }
      //riscaldo nuova copia name server
      riscalda();
      try {
         initFile();
      } catch( Exception e ) {
         if (DbFS.nameServer) 
            DbFS.outMexLn("(nameServer) Eccezione apertura file in scrittura "+e.toString());
      }
	}//nameServer
	
	public nameServer( String fileName ) throws FileNotFoundException {
      this( fileName, null );
   }//nameServer
	
	public nameServer( String fileName, gestore gs ) throws FileNotFoundException {
	   //ripristino di un nameServer precedentemente serializzato e salvato su file
	   this.gs = gs;
	   FileInputStream f = new FileInputStream( fileName+extNs );
	   name = fileName;
	   try {
   	   ObjectInputStream fObj = new ObjectInputStream( f );
   	   fisico = (Vector) fObj.readObject();
	      if (DbFS.nameServer) 
	         DbFS.outMexLn("(nameServer): (nameServer) deserializzato fisico");
   	   logico = (Vector) fObj.readObject();
	      if (DbFS.nameServer) 
	         DbFS.outMexLn("(nameServer): (nameServer) deserializzato logico\nLista"+
	         lista());
   	   nelem = ( (Integer) fObj.readObject() ).intValue();
	      if (DbFS.nameServer) 
	         DbFS.outMexLn("(nameServer): (nameServer) deserializzato nelem");
   	   f.close();
   	   fObj.close(); 
	   } catch( Exception e ) {
	      if (DbFS.nameServer) DbFS.outMexLn("(nameServer): (nameServer) Eccezione "+
	         e.toString() );
         fisico = new Vector( nameServer.dimDefault, nameServer.dimDefault );
         logico = new Vector( nameServer.dimDefault, nameServer.dimDefault );
         nelem = 0;
	   }
	   //riscaldo copia name server caricata
	   riscalda( fileName );
	   //ripristino prop file
      prop = new Vector( nameServer.dimDefault, nameServer.dimDefault );
	   elemProp eP;
	   try {
   	   f = new FileInputStream( fileName+extProp );
	      ObjectInputStream fObj = new ObjectInputStream( f );
	      try {
	         while( true ) {
	            //lettura proposta non confermata
   	         eP = ( elemProp ) fObj.readObject();
	            if ( !eP.valid ) 
	               //inserimento nel buffer della proposta
   	            prop.addElement( eP.r );
	            else {
	               //cancellazione dal buffer della proposta
	               int index = prop.indexOf( eP.r );
	               if ( index >= 0 ) prop.removeElementAt( index );
	            }
	         }
	      } catch( Exception e ) {
	         if (DbFS.nameServer) 
	            DbFS.outMexLn("(nameServer): Eccezione "+ e.toString() );
	      }
	   } catch( Exception e ) {
	      if (DbFS.nameServer) DbFS.outMexLn("(nameServer): (log) Eccezione "+
	         e.toString() );
	   }
	   //apertura file in scrittura   
      try {
         initFile( fileName );
      } catch( Exception e ) {
         if (DbFS.nameServer) 
            DbFS.outMexLn("(nameServer) Eccezione apertura file in scrittura "+e.toString());
      }
	}//nameServer
	
	protected void riscalda() {
	   riscalda( name );
	}//riscalda
	
	protected void riscalda ( String fileName ) {
      //aggiornamento copia fredda name server con log file
      ric r;
      risc = true;
	   try {
	      File fl = new File( fileName+extLog );
 	      if (DbFS.nameServer) 
 	         DbFS.outMexLn("(nameServer): (riscalda) fileLog: "+fl.getName());
	      if ( fl.exists() ) {
      	   fileLogRW = new RandomAccessFile( fileName+extLog, "rw" );
      	   FileInputStream f = new FileInputStream( fileLogRW.getFD() );
      	   ObjectInputStream fObj = new ObjectInputStream( f );
   	      if (DbFS.nameServer) 
   	         DbFS.outMexLn("(nameServer): (riscalda) agg. copia fredda (fileLog esiste)");
   	      try {
   	         while ( true ) {
      	         //lettura elemento
      	         r = ( ric ) fObj.readObject();
          	      if (DbFS.nameServer) 
      	            DbFS.outMexLn("(nameServer): (riscalda) letto "+ r.toString());
      	         if ( r.op ) 
      	            //inserimento in name server
      	            inserisci( r.fis, r.path );
      	         else
      	            //cancellazione da name server
      	            cancLogico( r.path );
   	         }
   	      } catch( Exception e ) {
   	      if (DbFS.nameServer) DbFS.outMexLn("(nameServer): (log) Eccezione "+
   	         "durante l'aggiornamento nameServer" + e.toString() );
   	      }
   	      finally {
   	         f.close();
   	         fObj.close();
   	         fileLogRW.close();
   	      }
	      }
 	   } catch( Exception e ) {
 	      if (DbFS.nameServer) DbFS.outMexLn("(nameServer): (log) Eccezione "+
 	         e.toString() );
 	   }	   
	   risc = false;
	}//riscalda
	
	protected void initFile() throws Exception {
	   initFile( nameDefault );
	}//initFile
	
	protected void initFile( String fileName ) throws Exception {
	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - ingresso");
	   //apertura file .log e .prp in scrittura
	   fileLogRW = new RandomAccessFile( fileName+extLog, "rw" );
	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - logRW");
	   fileLog = new FileOutputStream( fileLogRW.getFD() );
	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - log");
	   logObj = new ObjectOutputStream( fileLog );
	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - logObj");
	   if ( fileLogRW.length() > 0 )	   fileLogRW.seek( fileLogRW.length() );
	   filePropRW = new RandomAccessFile( fileName+extProp, "rw" );
	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - propRW");
	   fileProp = new FileOutputStream( filePropRW.getFD() ); 
	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - prop");
	   propObj = new ObjectOutputStream( fileProp );
	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - propObj");
      if ( filePropRW.length() > 0 )   filePropRW.seek( fileLogRW.length() );
	}//initFile 
	
	
	public void salva(  ) throws IOException  {
	   salva( name );
	}//salva
	
	public void salva( String fileName )  throws IOException {
	   //salvataggio su disco del nameServer
	   FileOutputStream f = new FileOutputStream( fileName+extNs );
	   try {
   	   ObjectOutputStream fObj = new ObjectOutputStream( f );
   	   fObj.writeObject( fisico );
	      if (DbFS.nameServer) 
	         DbFS.outMexLn("(nameServer): (salva) serializzato fisico");
   	   fObj.writeObject( logico );
	      if (DbFS.nameServer) 
	         DbFS.outMexLn("(nameServer): (salva) serializzato logico");
   	   fObj.writeObject( new Integer( nelem ) );
	      if (DbFS.nameServer) 
	         DbFS.outMexLn("(nameServer): (salva) serializzato nelem");
   	   f.close();
   	   fObj.close();
	      //punto di sincronizzazione: la copia su disco  coerente
	      //cancello log file
	      File fLog = new File( fileName+extLog );
	      if ( fLog.exists() ) {
	         //chiusura fLog
	         fileLogRW.close();
	         fileLog.close();
	         logObj.close();
   	      if (DbFS.nameServer) 
   	         DbFS.outMexLn("(nameServer): (salva) fLog esiste ("+
	            fLog.getAbsolutePath()+"): chiuso");
      	   File f2 = new File( "temp.txt" );
      	   boolean b = fLog.renameTo( f2 );
   	      if (DbFS.nameServer) 
   	         DbFS.outMexLn("(nameServer): (salva) fLog rename ("+f2.getName()+") = "+
	            String.valueOf( b ) );
      	   fileLogRW = new RandomAccessFile( fileName+extLog, "rw" );
      	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - logRW");
      	   fileLog = new FileOutputStream( fileLogRW.getFD() );
      	   if (DbFS.nameServer) DbFS.outMexLn("(name server): initFile - log");
      	   logObj = new ObjectOutputStream( fileLog );
   	      b = f2.delete();
   	      if (DbFS.nameServer) 
   	         DbFS.outMexLn("(nameServer): (salva) cancellato "+f2.getName()+" ( delete="+
   	         String.valueOf( b ) +")" );
	       }
	   } catch( Exception e ) {
	      if (DbFS.nameServer) 
	         DbFS.outMexLn("nameServer): (salva) Eccezione "+e.toString() );
	   }
	}//salva
	
	
	public static void newNSvuoto( String fileName ) throws IOException{
	   nameServer ns = new nameServer();
	   ns.salva( fileName );
	}//newNSvuoto

	public synchronized void inserisci( String fis, pathName log ) {
	   if (DbFS.nameServer) 
	      DbFS.outMexLn("(nameServer): inserisci( "+fis+", "+log.toString()+" )");
	   //controllo per evitare duplicazioni
	   int index = logico.indexOf( log );
	   if (DbFS.nameServer) 
	      DbFS.outMexLn("(nameServer): (inserisci) index("+log.toString()+")="+
	      String.valueOf( index ) );
	   if ( index < 0 ) {
   	   try {
      	   //scrittura sul log file
	         if (!risc) logObj.writeObject( new ric( log, fis, true ) );
            //inserimento nel name server in uso
            nelem++;
            fisico.addElement( fis );
            logico.addElement( log );
         } catch (Exception e) {
            if (DbFS.nameServer) {
               DbFS.outMexLn("(nameServer): inserisci Eccezione "+e.toString() );
               String sinc=new String("");
               synchronized ( sinc ) {
                 try { sinc.wait(3000); } catch( Exception f ) {}
               }
            }
            inserisci( fis, log );
         }
      }
	}//inserisci
	
	public synchronized void cancLogico( pathName log ) {
	   if (DbFS.nameServer) 
	      DbFS.outMexLn("(nameServer): cancLogico( "+log.toString()+" )");
		//ricerca indice elemento da cancellare
		int index = logico.indexOf( log );
	   if (DbFS.nameServer) 
	      DbFS.outMexLn("(nameServer): (cancLogico) index("+log.toString()+")="+
	      String.valueOf( index ) );
		try {
   		if ( index >= 0 ) {
   		   //scrittura sul log file
   		   if (!risc) logObj.writeObject( new ric( log, 
		                           ((String) fisico.elementAt( index )), false ));
   		   //cancellazione nel name Server in uso
      		nelem--;
      		fisico.removeElementAt( index );
      		logico.removeElementAt( index );
   		}
      } catch ( Exception e ) {
         if (DbFS.nameServer) 
            DbFS.outMexLn("(nameServer): cancLogico Eccezione "+e.toString() );
   	   cancLogico( log );
		}
	}//cancLogico
	
	public synchronized void cancFisico( String fis ) {
		//ricerca indice elemento da cancellare
		int index = fisico.indexOf( fis );
		try {
   		if ( index >= 0 ) {
   		   //scrittura sul log file
   		   if (!risc) logObj.writeObject( 
		            new ric( ((pathName) logico.elementAt( index )), fis, false ));
   		   //cancellazione nel name Server in uso
      		nelem--;
      		fisico.removeElementAt( index );
      		logico.removeElementAt( index );
   		}
		} catch (Exception e) {
		   cancFisico( fis );
		}
	}//cancFisico
	
	public synchronized String ricLogico( String fis ) {
		int index = fisico.indexOf( fis );
		if (index >= 0 ) return ((pathName) logico.elementAt( index )).toString();
		return pathName.pathNull.toString();
	}//ricLogico
	
	public synchronized String ricFisico( pathName log ) {
		int index = logico.indexOf( log );
		if (index >= 0 ) return ( (String) fisico.elementAt( index ) );
		return nameServer.StringNull;
	}//ricFisico
	
	public synchronized boolean insProp( ric r ) {
	//inserimento nuova proposta
	   try {
	      int index = prop.indexOf( r );
	      if ( index < 0 ) {
     	      //scrittura sul file delle proposte
   	      propObj.writeObject( new elemProp( r, false ) );
   	      //scrittura sul buffer in uso
   	      prop.addElement( r );
	         return true;
	      }
	      else return false;
	   } catch (Exception e) {
	      return insProp( r );
	   }
	}//insProp
	
	public synchronized void cancProp( ric r ) {
	//cancellazione proposta
	   try {
	      int index = prop.indexOf( r );
	      if ( index >= 0 ) {
     	      //scrittura sul file delle proposte
   	      propObj.writeObject( new elemProp( r, true ) );
   	      //scrittura sul buffer in uso
   	      prop.removeElementAt( index );
	      }
	   } catch ( Exception e ) {
	      cancProp( r );
	   } 
	}//cancProp
	
	public synchronized ric ricProp( ric r ) {
	//ricerca all'interno della tabella delle proposta una che interessi lo stesso
	//nome logico di r.log
	   int index = prop.indexOf( r );
	   if ( index>=0 ) return ( (ric) prop.elementAt( index ) );
	   else return ric.ricNull;
	}//ricProp
	
	public synchronized String[] listaLogico() {
		String lista[] = new String[ nelem ];
		int index = 0;
		int indArray = 0;
		pathName path;
		//scansione della tabella e costruzione dell'array
		while( index < logico.size() ) {
			path = (pathName) logico.elementAt( index );
			lista[indArray] = path.toString();
			indArray++;
			index++;
		}
		return lista;
	}//listaLog
	
	public synchronized String[] listaFisico() {
		String lista[] = new String[ nelem ];
		int index = 0;
		int indArray = 0;
		String path;
		//scansione della tabella e costruzione dell'array
		while( index < fisico.size() ) {
			path = (String) fisico.elementAt( index );
			lista[indArray] = path;
			indArray++;
			index++;
		}
		return lista;
	}//listaFis
	
	public synchronized String lista() {
		String out = new String ("Name Server:\n");
		out = out.concat("( <nome logico> . <nome fisico> )\n");
		int index = 0;
		pathName path;
		String pathFisico;
		//scansione della tabella 
		while( index < fisico.size() ) {
			pathFisico = (String) fisico.elementAt( index );
			path = (pathName) logico.elementAt( index );
			out = out.concat( " ( " + path.toString() + " . " );
			out = out.concat( pathFisico + " )\n" );
			index++;
		}
		out = out.concat( "Name Server: fine lista\n" );
		return out;
	}//lista
	
	public synchronized String listaProp() {
		int index = 0;
		String out = new String("");
		ric r;
		//scansione della tabella 
		while( index < prop.size() ) {
		   r = (ric) prop.elementAt( index );
			out = out.concat( String.valueOf( index) + r.toString() + "\n" );
			index++;
		}
		out = out.concat( " \n " );
		return out;
	}//lista
	
   boolean isValid( char ch ) {
      if ( '0'<= ch && ch <= '9' || ch=='.' ) return true;
         else return false;
   }//isValid
	
	public long f( InetAddress add, pathName path ) {
   	try {
   	   int molt = -1;
   	   if ( path.toString().length() % 2 == 0 ) molt = 1;
   	   String IP = add.getHostAddress();
   	   StringReader st = new StringReader( IP );
   	   char c[] = new char[3];
   	   long num = 0;
   	   for (int i=0; i<4; i++) {
      	   c[0] = c[1] = c[2] = '0';
      	   int index = 0;
      	   char ch = '0';
      	   while (ch != '.'&& index<3 && isValid(ch)) {
      	      ch = (char )st.read();
      	      if (ch!='.'&& isValid(ch)) {
      	         c[0] = c[1];
      	         c[1] = c[2];
      	         c[2] = ch;
      	      }
   	         index++;
      	   }
   	      if (index==3) st.read();
   	      num = num * 1000 + Integer.valueOf( new String( c ) ).longValue() ;
   	   }
   	   return num*molt;
   	}
   	catch (Exception e) {
   	   if (DbFS.nameServer) DbFS.outMexLn("f: Eccezione "+e.toString() );
   	   return f( add, path );
   	}
	}//f
	
	public long f( InetAddress add, int inPort, pathName path ) { 
	   return f( add, path) + inPort;
	}//f
	
}//nameServer

class nameServerRid implements Serializable {
//crea il name server da esportare ai nuovi nodi

   Vector fisico;
   Vector logico;
   int nelem;
   Vector prop;
   
   public nameServerRid( nameServer ns ) {
	   logico = (Vector) ns.logico.clone();
	   fisico = (Vector) ns.fisico.clone();
	   for ( int i=0; i<fisico.size(); i++ )
	      if ( !((String) fisico.elementAt(i)).equals( nameServer.fisicoEsterno )) 
	         fisico.setElementAt( nameServer.fisicoEsterno, i );
	   nelem = ns.nelem;
	   prop = new Vector( nameServer.dimDefault, nameServer.dimDefault );
   }//nameServerRid

}//nameServerRid

