//-----------------------------------------------------------------------------
// predScelto.java
// package: FileServer
// 
// comando con cui si notifica ad un nodo di essere il predecessore di un nuovo 
// nodo che si vuole inserire
//
// data: 27-02-2000
//-----------------------------------------------------------------------------

package FileServer;

import java.io.*;
import java.net.*;
import java.util.*;

public class predScelto extends Comando{

   gestReteAnello gR;
   InetAddress newServer;
   int portNewServ;

   public predScelto( InetAddress newServer, int portNewServ ) {
      super( "predScelto" );
      this.newServer = newServer; 
      this.portNewServ = portNewServ;  
   }//predScelto

   public void set( gestore gR ) {
      this.gR = (gestReteAnello) gR;
   }//set
   
   public void run() {
      try {
         if ( gR.sockCom != null ) gR.sockCom.close();
         if (DbFS.cmdRete) DbFS.outMexLn("(predScelto): chiusa socket ricezione comando");
         //aggiornamento nuovo server
         //creazione canale di comunicazione con il nuovo server
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): tentativo di aprire la socket con il "+
            "nuovo server( IP="+newServer+";port="+String.valueOf(portNewServ) );
         Socket sAgg = new Socket( newServer, portNewServ );
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): aperta la socket");
         //aggiornamento
         boolean oldDisp = gR.disp;
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): aggiornamento nuovo server");
         gR.operServ( sAgg, newServer, portNewServ );
         //comunicazione successore (confermata) ( comando di aggiornamento )
         Socket sSucc = null;
         if (oldDisp) {
            //successore presente
            sSucc = new Socket( gR.succAdd, gR.portServSucc );
            if (DbFS.cmdRete) 
               DbFS.outMexLn("(predScelto): creata connessione con successore");
            ObjectOutputStream outSucc = new ObjectOutputStream( sSucc.getOutputStream() );
            outSucc.writeObject( new newSucc( newServer, portNewServ ) );
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): inviato comando aggiornamento a successore");
         }
         //comunicazione predecessore (confermata)(comando di aggiornamento )
         //indirizzo predecessore attuale
         InetAddress predAct = ((Socket) gR.inSock.elementAt( 0 )).getInetAddress();
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): predecessore attuale "+predAct);
         Socket sPred;
         if ( predAct.equals( gR.pred ) )
            //predecessore  presente
            sPred = new Socket( gR.pred, gR.portServPred );
         else
            //predecessore caduto
            sPred = new Socket( predAct, gR.portServPred2 );
         ObjectOutputStream outPred = new ObjectOutputStream( sPred.getOutputStream() );
         outPred.writeObject( new newPred( newServer, gR.inPortNS ) );
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): inviato comando aggiornamento a predecessore");
         //attesa risposta succ e pred (sequenziale)
         if (oldDisp) {
            ObjectInputStream inSucc = new ObjectInputStream( sSucc.getInputStream() );
            try {
               inSucc.readObject();
            } catch( Exception e ) {
               if (DbFS.cmdRete) 
                  DbFS.outMexLn("(predScelto): Eccezione lettura ack "+e.toString());
            }
            if (DbFS.cmdRete) 
               DbFS.outMexLn("(predScelto): letto ack da predecessore");
            sSucc.close();
         }
         ObjectInputStream inPred = new ObjectInputStream( sPred.getInputStream() );
         inPred.readObject();
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): letto ack da successore");
         sPred.close();
         //nodi predecessore e successore aggiornati
         //aggiornamento nodo attuale
         gR.succ2Add = gR.succAdd;
         gR.succ2Port = gR.succPort;
         gR.succAdd = newServer;
         gR.succPort = gR.inPortNS;
         gR.portServSucc = portNewServ;
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): aggiornato nodo");
         if (DbFS.cmdRete) DbFS.outMexLn("(predScelto): salvataggio configurazione ("+
            gR.fileConfig+")" );
         gR.saveConnOut( gR.fileConfig );
         //cancella canale server con successore attuale
         //da qui in poi il meccanismo si automatizza
         Socket sAtt = (Socket) gR.outSock.elementAt( 0 );
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): chiuso collegamento con succ attuale");
         sAtt.close();
         //invia ack a nuovo nodo per dare il via ripristino del successore
         //chiusura socket
         sAgg.close();
         //crea canale server con nuovo successore (nuovo server)
         //questo dovrebbe farlo da solo
      }
      catch ( Exception e ) {
         if (DbFS.cmdRete) 
            DbFS.outMexLn("(predScelto): Eccezione "+e.toString() );
      }
   }//run
   
}//predScelto